/* Emperor - an X windows strategy game for 1-6 players
   Copyright (C) 1998 Frank Rustemeyer.
   This file is part of the Emperor package.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

   If you have changes and you want them to be included in a future
   release of Emperor, please e-mail them to mambres@mud.tap.de.

   This file implements the game logic. It is (and should remain to be)
   independent of the visualisation.

   98/11/30, Frank Rustemeyer <mambres@mud.tap.de>: Released.
   99/01/29, Frank Rustemeyer <mambres@mud.tap.de>: Added roads.
   99/01/29, Frank Rustemeyer <mambres@mud.tap.de>: Added language support.
   99/02/09, Frank Rustemeyer <mambres@mud.tap.de>: Fixed and straightened the computer move
   99/02/11, Frank Rustemeyer <mambres@mud.tap.de>: Major rewrite of the computer player
   99/04/05, Frank Rustemeyer <mambres@mud.tap.de>: Bugfix: handle save files without units
*/

#include "emp.h"
#include "empgame.h"

#include <string.h>
//  strcpy in spieler::spieler
//  strncmp in list::load
#include <stdlib.h>
//  rand() in RANDOM(n) in map::growth, map::map
#include <time.h>
//  randomize() in map::map
//#include <stdio.h>  included in empgame.h
//  sprintf() in unit::restmovestring, flieger::restmovestring
//  FILE, fopen(), fclose(), fprintf() in the save routines

#define LANDTARGETS 1
#define WATERTARGETS 0
#define NAVALTARGETS 2
#define AIRTARGETS 3
#define HOMECITIES 4

extern game *g;

//*********************************************************** list ****

list::list(void)      // create an empty list
{   root = NULL;
    len = 0;
    pos = 0;
    }

int list::init(class listmember * member)
  // if list has been empty when member is added, initialize list,
  // otherwise return zero.
{   if (root) return 0;   // return if list has not been empty
    root = member;    // this member will now be the root
    pos = 1;
    return 1;
    }

int list::before(class listmember * member) const
  // does member appear in the list before root ?
{   class listmember * seek = root->previous();
      // start with one before root
    while(seek)                          // as long as there are members
    {   if (seek == member) return 1;    // is it this one ? yes = return
	seek = seek->previous();         // no: one further back
	}
    return 0;  // member was not among the members tested: return false
    }

void list::remove(class listmember * member)
  // removes member from the list
{   if (root == member)         // if root points to member
	if (member->next())     //   and is not at the end of the list
	    root = member->next();
	      // root will now point to the next member
	else root = member->previous();
	  // if it is end of the list, root will point to the previous
    len--;                      // list will no be one shorter
    member->unlink();
    if (member) delete member;                
    }

listmember *list::find(coordinate c)
{   if (!root) return NULL;
    return root->first()->find(c);
    }


void list::save(FILE *f)
  // save this list.
{   listmember *run;

    run = NULL;
    if (root) run = root->first();    // start with first list member
    while(run)              // as long as there are members
    {   run->save(f);       //   save this one
	run = run->next();  //   next member
	}
    fprintf(f, "END\n");
    }

void list::load(FILE *f, int flag)
  // load this list. 
{   char s[100];
    listmember *run;
    int utype;

    while(root)
        remove(root);

    while (fgets(s, 100, f) && (s[0] != 'E'))
    {   if (flag == CITIES) run = new city();
        else
        {   sscanf(s, "%d", &utype);
            fgets(s, 100, f);
            switch(utype)
            {    case INFANTRIE:
                      run = new infantrie();
                      break;
                 case PANZER:
                      run = new panzer();
                      break;

                 case FLIEGER:
	              run = new flieger();
                      break;

                 case TRANSPORTER:
	              run = new transporter();
	              break;

	         case UBOOT:
	              run = new uboot();
	              break;

	         case SCHNELLBOOT:
	              run = new schnellboot();
	              break;

	         case FREGATTE:
	              run = new fregatte();
	              break;

	         case SCHLACHTSCHIFF:
	              run = new schlachtschiff();
	              break;

	         case MINENLEGER:
	              run = new minenleger();
                      break;
                 default: run = new infantrie(); //shouldn't happen
                 }
            }
	run->load(s);
        run->addtolist(*this);
	  // parse data from the buffer string into the list member
        }
    }

// **************************************************** listmember ****

void listmember::addtolist(list &thislist, listmember *aft)
  // insert this object into thislist after member aft, or at end of list
{   prv = nxt = NULL;
    thislist.len++;          // thislist will now be one longer
    if (thislist.init(this)) // initialize list in case it was empty
	return;
    if (!aft) aft = thislist.root->last();
      // append at end of list if aft not specified
    nxt = aft->nxt;   // put new member in pointer chain
    aft->nxt = this;
    prv = aft;
    if (nxt) nxt->prv = this;
    }

listmember::listmember(list &thislist, listmember *aft)
  // create a new member in thislist after member aft, or at end of list
{   addtolist(thislist, aft);
    }

listmember::listmember(void)
  // alternate constructor: listmember outside a list
{   prv = nxt = NULL;
    }

//listmember::~listmember(void)
  // destruct this list member (only to be called from list::remove)
void listmember::unlink(void)
{   if (prv) prv->nxt = nxt;
    if (nxt) nxt->prv = prv;
     // let previous and next member point to each other
    }

listmember * listmember::first(void) const
  // return the first list member:
{   const listmember * seek = this;      // start here,
    while (seek->prv) seek = seek->prv;
      // and go backwards as long as possible
    return (listmember *) seek;
      // cast is necessary to make pointer not const
    }

listmember * listmember::last(void) const
  // return the last member in the list:
{   const listmember * seek = this;      // start here,
    while (seek->nxt) seek = seek->nxt;
      // and go forward as long as possible
    return (listmember *) seek;
      // cast is necessary to make pointer not const
    }

listmember * listmember::find(coordinate c) const
{   if (pos == c) return (listmember *) this; // make unconst
    if (!nxt) return NULL;
    return nxt->find(c);
    }

void listmember::setowner(char o)
{   owner = o;
    g->ownership[pos] = o;
    g->explore(pos, o);
    }

//*********************************************** coordinate **********

coordinate coordinate::operator + (coordinate &c)
{   coordinate r;
    r.set(x+c.x, y+c.y);
    return r;
    }

int coordinate::operator - (coordinate &c)
{   int dx, dy;

    dx = x - c.x;
    if (dx < 0) dx = -dx;
    dy = y - c.y;
    if (dy < 0) dy = -dy;

    return (dx > dy) ? dx : dy;
    }

int coordinate::operator == (const coordinate &c) const
{   return (x == c.x) && (y == c.y);
    }

int coordinate::operator != (const coordinate &c) const 
{   return (x != c.x) || (y != c.y);
    }

//******************************************************* matrix ******

matrix::matrix()
{   coordinate i;
    LOOP(i, 0,0, SX-1,SY-1, 1)
        (*this)[i] = 0;
    }

void matrix::save(FILE *fl)
{   int i;
    for (i = 0; i < SX*SY; i+=10)
       fprintf(fl, "%d %d %d %d %d %d %d %d %d %d\n",
           f[i], f[i+1], f[i+2], f[i+3], f[i+4], f[i+5], f[i+6], f[i+7], f[i+8], f[i+9]);
    }

void matrix::load(FILE *fl)
{   int i;
    char s[100];
    for (i = 0; i < SX*SY; i+=10)
    {   fgets(s, 100, fl);
        sscanf(s, "%d %d %d %d %d %d %d %d %d %d",
           &f[i], &f[i+1], &f[i+2], &f[i+3], &f[i+4], &f[i+5], &f[i+6], &f[i+7], &f[i+8], &f[i+9]);
        }
    }

//********************************************************* map *******

void map::growth(coordinate c)
{   coordinate i;

    LOOP(i, c.x-1, c.y-1, c.x+1, c.y+1, 1)
        if (ONMAP(i) && !(*this)[i] && (RANDOM(100) < 19))
        {   if (RANDOM(100) < 4)
	        (*this)[i] = 5;
            else (*this)[i] = 3 + RANDOM(2);
	        growth(i);
            }
    }

map::map(void) : matrix()
{   coordinate i,r;
  
    RANDOMIZE();
    LOOP(i, 0,0, SX-1,SY-1, 10)
    {   r.set(RANDOM(10) +i.x, RANDOM(10) +i.y);
        if (!(*this)[r]) (*this)[r] = 3 + RANDOM(2);
	growth(r);
	}
    LOOP(i, 0,0, SX-1, SY-1, 1)
    {   if ((*this)[i]>2)
            LOOP(r, i.x-1, i.y-1, i.x+1, i.y+1, 1)
		        if (ONMAP(r) && ((*this)[r]<2))
		               (*this)[r] = 2;
	        if (!(*this)[i]) (*this)[i]+=RANDOM(2);
	    }
    }

//****************************************************** city *********

city::city(list &thislist) : listmember(thislist)
{   owner = 0;
    pos.set(0,0);
    ptime = 0;
    produced = NULL;
    }

city::city() : listmember()
{   owner = 0;
    pos.set(0,0);
    ptime = 0;
    produced = NULL;
    }

/* City production. Return number of cities owned. */
int city::produce(char o)
{   int typ;
    if ((owner == o) && produced)
    {   ptime++;
        if (ptime == produced -> prodtime())
        {   ptime = 2;
            produced -> setlfdnr((g->unitcount)++);
            produced -> setowner(o);
            produced -> setpos(getpos());
            produced -> addtolist(g->units);
            typ = produced->id();
            setprod(typ);
            }
        }
    if (next()) return ((city *) next()) -> produce(o) + (owner == o ? 1 : 0);
    return (owner == o ? 1 : 0);
    }

void city::setprod(int typ)
{   if (produced && !(produced->getlfdnr()) && (produced->id() == typ)) return;
    ptime = 2;
    if (!produced || (produced->id() != typ)) ptime = 0; 
    if (produced && !(produced->getlfdnr())) delete produced;
    produced = NULL;
    switch(typ)
    {   case EXIT:
	     return;

	case INFANTRIE:
	     produced = new infantrie();
	     return;

	case PANZER:
	     produced = new panzer();
	     return;

	case FLIEGER:
	     produced = new flieger();
	     return;

	case TRANSPORTER:
	     produced = new transporter();
	     return;

	case UBOOT:
	     produced = new uboot();
	     return;

	case SCHNELLBOOT:
	     produced = new schnellboot();
	     return;

	case FREGATTE:
	     produced = new fregatte();
	     return;

	case SCHLACHTSCHIFF:
	     produced = new schlachtschiff();
	     return;

	case MINENLEGER:
	     produced = new minenleger();
	     return;

	    }
   }

void city::save(FILE *f)
{   fprintf(f, "%d %d %d %d %d\n", pos.x, pos.y, owner, ptime, (produced ? produced->id():0));
    }

#pragma argsused
void city::load(char *s)
{   int loadprod, time;
    sscanf(s, "%d %d %d %d %d", &(pos.x), &(pos.y), &owner, &time, &loadprod);
    if (loadprod) setprod(loadprod);
    else produced = NULL;
    ptime = time;
    g->ownership[pos] = owner;
    }

//******************************************************* unit *******

unit::unit() : listmember()
{   pos.set(-1,-1);
    owner = 0;
    damage = 0;
    carried = 0;
    repaired = 0;
    lfdnr = 0;
    }

unit::unit(list &thislist) : listmember(thislist)
{   pos.set(-1,-1);
    owner = 0;
    damage = 0;
    carried = 0;
    repaired = 0;
    lfdnr = 0;
    }

void unit::setpos(coordinate c)
{   unit * other;
    coordinate oldpos;
    oldpos = pos;
    pos.set(c.x, c.y);
    if ((oldpos.x != -1) && (oldpos.y != -1)) 
    {  if (!g->stadt(oldpos)) g->ownership[oldpos] = 0;
       if (FIRSTMEMBER(g->units))
          other = (unit *) g->units.find(oldpos);
       if (other) g->unitmap[oldpos] = other->id();
       else g->unitmap[oldpos] = 0;
       }
    if ((c.x == -1) || (c.y == -1)) return;
    g->unitmap[pos] = id();
    g->ownership[pos] = getowner();
    if (owner) g->explore(pos, getowner());
    if ((id() == FLIEGER) || (id() == UBOOT)) return;
    if ((oldpos.x == -1) || (oldpos.y == -1)) return; //avoid hitting mines when loading game
    if (g->vermint(pos) && (RANDOM(100) < 50)) hit();
    };

char * unit::prodtimestring(char * s)
{   itoa(prodtime(), s, 10);
    return s;
    }

char * unit::restmovestring(char *s)
{   sprintf(s, "%d ", restmoves());
    return s;
    }

char * unit::hitpointstring(char * s)
{   sprintf(s, T_STRENGTH, name(), maxdamage()-damage, maxdamage());
    return s;
    }

coordinate unit::movetowards(coordinate m)
{   int moving = 1;
    coordinate nextsquare;
    
    while(!(m == pos) && (damage < maxdamage()) && moving)
    {   nextsquare = pos;
        if (m.x > pos.x) nextsquare.x ++;
        if (m.x < pos.x) nextsquare.x --;
        if (m.y > pos.y) nextsquare.y ++;
        if (m.y < pos.y) nextsquare.y --;
        moving = validsquare(nextsquare);
        if (moving == 2) return nextsquare; // transporter unload 
        if (moving && (hasmoved < canmove()))
        {   setpos(nextsquare);
            hasmoved += movement_cost(nextsquare);
            }
        else moving = 0; // all movement points used
        }
    return pos;
    }

char * unit::repair(char * s)
{   if (!damage)
    {   strcpy(s, T_HEALTHY);
        return s;
        }
    if (++repaired == repairtime())
    {   repaired = 0;
        damage--;
        if (damage) strcpy(s, T_PARTIAL);
        else strcpy(s, T_COMPLETE);
        return s;
        }
    if (repairtime()-repaired == 1) strcpy(s, T_ONEROUND);
                           // fr Uli :)
    else sprintf(s, T_ROUNDSTOGO, repairtime()-repaired);
    return s;
    }

void unit::kill(void)
{   unit * loaded, * other;
    coordinate oldpos;

    while (FIRSTMEMBER(g->units) && (loaded = ((unit *) FIRSTMEMBER(g->units))->findcarried(getlfdnr())))
    {   loaded -> kill();
        }
    oldpos = pos;
    pos.set(-1, -1); // hide this unit from find()
    if ((oldpos.x != -1) && (oldpos.y != -1))
    {   other = (unit *) g->units.find(oldpos);
        if (other) g->unitmap[oldpos] = other->id();
        else
        {   g->unitmap[oldpos] = 0;
            if (!g->stadt(oldpos)) g->ownership[oldpos] = 0;
            }
        }
    g->units.remove((listmember *) this);
    }

class unit *unit::alive(void)
{   if (damage < maxdamage()) return this;
    kill();

    return NULL;
    }

int unit::checkcombat(coordinate c)
{   unit *enemy;
    if ((*(g->mp))[c] == 5) return 1;
    enemy = (unit *) g->units.find(c);
    if (!enemy) return 1;
    if (enemy->getowner() == getowner()) return 0;

    while (enemy->alive() && (damage < maxdamage()))
    {   if (RANDOM(100) < 50) hit();
        if (RANDOM(100) < 50) enemy->hit();
        }
    if (damage < maxdamage()) return 1;
    return 0;
    }

/* clear all moves and count own units */
int unit::clearmoves(int o)
{   hasmoved = 0;
    if (next()) return ((unit *) next())->clearmoves(o) + (owner == o ? 1 : 0);
    return (owner == o ? 1: 0);
    }

class unit * unit::findnext(int player , int start)
{   class unit *weiter;
    if ((owner == player) && (!carried) && restmoves()) return this;
    weiter = (unit *) next();
    if (!weiter)
    {   if (!start) return NULL;
        weiter = (unit *) first();
        }
    if (weiter->getlfdnr() == start) return NULL;
    return weiter->findnext(player, start);
    }

class unit * unit::findcarried(int transp)
{   if (carried == transp) return this;
    if (!next()) return NULL;
    return ((unit *) next())->findcarried(transp);
    }

class unit * unit::findlfdnr(int nr)
{   if (nr == lfdnr) return this;
    if (!next()) return NULL;
    return ((unit *) next())->findlfdnr(nr);
    }

void unit::save(FILE *f)
{   fprintf(f, "%d\n%d %d %d %d %d %d %d %d\n", id(), lfdnr, pos.x, pos.y, getowner(), hasmoved, damage, repaired, carried);
    }

void unit::load(char *s)
{   coordinate c;
    sscanf(s, "%d %d %d %d %d %d %d %d", &lfdnr, &c.x, &c.y, &owner, &hasmoved, &damage, &repaired, &carried);
    setpos(c);
    }

//**************************************************** landunit ******

int landunit::validsquare(coordinate c)
{   city *ct;
    transporter *transp;
    unit *enemy;
    coordinate off;
    off.set(-1, -1);
    if (!ONMAP(c)) return 0;
    if (hasmoved >= canmove()) return 0;
    if (g->wasser(c))
    {   transp = (transporter *) g->units.find(c);
        if (!transp) return 0;
        if (transp->getowner() != getowner()) return 0;
        carried = transp->getlfdnr();
        transp->setnextunload(getlfdnr());
        setpos(off);
        return 0;
        }
    if (g->stadt(c))
    {   ct = (city *) g->cities.find(c);
        if (ct->getowner() == getowner()) return 1;  // own city
        if (RANDOM(100) < 50)   // failure
        {   hasmoved++;
            this->hit();
            return 0;
            }
        damage = maxdamage(); // destroy conquering unit
        while(enemy = (unit *) g->units.find(c)) enemy->kill(); //destroy conquered units
        ct->setowner(getowner());
        ct->setprod(EXIT);  // cancel production        
        }

    return checkcombat(c);
    }

int landunit::movement_cost(coordinate destination)
{   if (g->strasse(destination)) return 0;
    return 1;
    }
 
//**************************************************** flieger *******

int flieger::validsquare(coordinate c)
{   city *ct;
    int free;

    if (!ONMAP(c)) return 0;
    if (hasmoved >= canmove()) return 0;
    if (g->stadt(c))
    {   ct = (city *) g->cities.find(c);
        if (ct->getowner() != getowner()) return 0;  // enemy city
        fuel = 21;
        }
    free = checkcombat(c);
    if (free && !--fuel) hit();
    return free;
    }

char * flieger::restmovestring(char *s)
{   sprintf(s, "%d/%d ", restmoves(), fuel);
    return s;
    }

void flieger::save(FILE *f)
{   fprintf(f, "%d\n%d %d %d %d %d %d %d %d %d\n", id(), lfdnr, pos.x, pos.y, getowner(), hasmoved, damage, repaired, carried, fuel);
    }

void flieger::load(char *s)
{   coordinate c;
    sscanf(s, "%d %d %d %d %d %d %d %d %d", &lfdnr, &c.x, &c.y, &owner, &hasmoved, &damage, &repaired, &carried, &fuel);
    setpos(c);
    }

//**************************************************** ship **********

int ship::validsquare(coordinate c)
{   city *ct;
    if (!ONMAP(c)) return 0;
    if (hasmoved >= canmove()) return 0;
    if (g->stadt(c))
    {   ct = (city *) g->cities.find(c);
        if (ct->getowner() != getowner()) return 0;  // enemy city
        }
    else if ((*(g->mp))[c] > 2) return 0;
    return checkcombat(c);
    }

//**************************************************** transporter ***

int transporter::validsquare(coordinate c)
{   unit *unload;
    if (!ONMAP(c)) return 0;
    if (!g->wasser(c) && !g->stadt(c))
    {   if (!nextunload) return 0;
        if (g->units.find(c)) return 0;
        if (!FIRSTMEMBER(g->units)) return 0;
        unload = ((unit *)FIRSTMEMBER(g->units))->findlfdnr(nextunload);
        if (!unload) return 0;
        unload->setpos(c);
        unload->countmove();
        unload->setcarried(0);
        nextunload = 0;
        if (FIRSTMEMBER(g->units))
        {   unload = ((unit *) FIRSTMEMBER(g->units))->findcarried(getlfdnr());
            if (unload) nextunload = unload->getlfdnr();
            }
        return 2;
        }
    return ship::validsquare(c);
    }

void transporter::save(FILE *f)
{   fprintf(f, "%d\n%d %d %d %d %d %d %d %d %d\n", id(), lfdnr, pos.x, pos.y, getowner(), hasmoved, damage, repaired, carried, nextunload);
    }

void transporter::load(char *s)
{   coordinate c;
    sscanf(s, "%d %d %d %d %d %d %d %d %d", &lfdnr, &c.x, &c.y, &owner, &hasmoved, &damage, &repaired, &carried, &nextunload);
    setpos(c);
    }

//**************************************************** spieler *******

spieler::spieler()
{   strcpy(name, "Fred");
    used = 0;
    computer = 0;
    }

void spieler::save(FILE *f)
{   fprintf(f, "%d %d %s\n", used, computer, name);
    }

void spieler::load(FILE *f)
{   char s[100];
    fgets(s, 100, f);
    sscanf(s, "%d %d %s", &used, &computer, name);
    }

//******************************************************* game *******

game::game() : 
      cities(), units(), ownership(), unitmap(), mine(), known(), seen()
{   activeplayer = 0;
    unitcount = 1;
    strcpy(sp[0].name, T_NEUTRAL);
    strcpy(sp[1].name, T_FRIEDRICH);
    strcpy(sp[2].name, T_DONEGAL);
    strcpy(sp[3].name, T_DIEGO);
    strcpy(sp[4].name, T_CARL);
    strcpy(sp[5].name, T_THORLEIF);
    strcpy(sp[6].name, T_CHARLES);
    }

void game::explore(coordinate c, int spi)
{   coordinate i;
    if ((c.x == -1) || (c.y == -1)) return;
    LOOP(i, c.x-1, c.y-1, c.x+1, c.y+1, 1)
        if (ONMAP(i))
        {   known[i] |= (1 << spi);
            seen[i] = 1;
            }
    }

void game::lookout(int sp)
{   coordinate i, c;
    int dir;
    LOOP(i, 0, 0, SX-1, SY-1, 1) 
    {  seen[i] = 0;
       if ((*mp)[i] == 6)
       {   dir = 1; roadmap[i] = 0;
           LOOP(c, i.x-1, i.y-1, i.x+1, i.y+1, 1)
               if (!(c == i))
               {   if (stadt(c) || strasse(c)) roadmap[i] += dir;
                   dir *= 2;
                   }
           if (!(roadmap[i])) roadmap[i] = 2; 
           }
       }
    LOOP(i, 0,0, SX-1, SY-1, 1)
       if (ownership[i] == sp) explore(i, sp);
    }

void game::initcities(void)
{   int spl, i, min, dist;
    coordinate centre;
    city *c, *found;
    infantrie *inf;
   
    LOOP(centre, 0, 0, SX, SY, 1)
        if (stadt(centre))
        {   c = new city(cities);
            c->setpos(centre);
            }

    for(spl = 1; spl < 7; spl++)
        if (g->sp[spl].used)
	{   switch(spl)
	    {   case 1: centre.set(0,0);
			break;
		case 2: centre.set(SX, SY);
			break;
		case 3: centre.set(0, SY);
			break;
		case 4: centre.set(SX, 0);
			break;
		case 5: centre.set(SX / 2, 0);
			break;
		case 6: centre.set(SX / 2, SY);
			break;
		default:
			centre.set(0,0); //should not happen
		}

	    for(i=0; i < BEGINNSTAEDTE; i++)
    	    {   min = SX+SY+1;
		c = (city *) FIRSTMEMBER(cities);
                found = NULL;

		while(c)
		{   if (!c->getowner())
		    {   dist = c->getpos() - centre;
			if (dist < min)
			{   min = dist;
			    found = c;
			    }
			}
		    c = (city *) c->next();
		    }

		if (found)
		{   found->setowner(spl);
		    inf = new infantrie(g->units);
                    inf->setlfdnr(unitcount++);
		    inf->setowner(spl);
		    inf->setpos(found->getpos());
		    }
		}
	    }
    }    

int game::load(char *file)
{   FILE *f;
    int i;
    char s[100];
    unit *run;
    city *ct;
    coordinate c;

    LOOP(c, 0, 0, SX-1, SY-1, 1)
    {   ownership[c] = unitmap[c] = 0;
        }
    ct = (city *) FIRSTMEMBER(g->cities);
    while(ct)
    {   if (ct->produced) delete ct->produced;
        ct = (city *) ct->next();
        }

    f = fopen(file, "r");
    if (!f) return 0;

    fgets(s, 100, f);
    fgets(s, 100, f);
    sscanf(s, "%d ", &activeplayer);
    for (i = 0; i<7; i++) sp[i].load(f);

    mp->load(f);
    mine.load(f);
    known.load(f);

    cities.load(f, CITIES);
    units.load(f, ARMIES);

    if (fclose(f)) return 0;

    unitcount = 0;
    run = (unit *) FIRSTMEMBER(units);
    while(run)
    {   if (run->getlfdnr() > unitcount) unitcount = run->getlfdnr();
        run = (unit *) run->next();
        }
    unitcount++;
    return 1;
    }

int game::save(char *file)
{    FILE *f;
     int i;

     f = fopen(file, "w");
     if (!f) return 0;
     fprintf(f, "Emperor Savefile v1.1\n%d \n", activeplayer);
     for (i = 0; i<7; i++) sp[i].save(f);
     mp->save(f);
     mine.save(f);
     known.save(f);
     cities.save(f);
     units.save(f);
     if (fclose(f)) return 0;
     return 1;
     }

int game::vermint(coordinate c)
{   if (!mine[c]) return 0;
    mine[c] |= (1 << activeplayer);  // Mine dem Spieler bekannt machen
    return 1;
    }

int game::minebekannt(coordinate c)
{   return mine[c] & (1 << activeplayer);
    }

void game::minelegen(coordinate c)
{   mine[c] |= (1 << activeplayer);
    }

int game::bekannt(coordinate c)
{   return erkundet(c);
    }

int game::bekannt(coordinate c, int sp)
{   return known[c] &(1 << sp);
    }

int game::beobachtet(coordinate c)
{   return seen[c];
    }

int game::beobachtet(coordinate c, int sp)
{   return seen[c];
    }

int game::coast(coordinate c)
{   coordinate i;

    LOOP(i, c.x -1, c.y -1, c.x +1, c.y +1, 1)
        if (ONMAP(i) && wasser(i)) return 1;
    return 0; 
    }

void game::compmove(void)
{   int move_land, move_ferries, move_ships, move_air;
    unit *u, *nextu;
    plan *land;

    u = (unit *) FIRSTMEMBER(units);
    while (u)
    {    if (u->getowner(1) == activeplayer)
         {   if (u->is_landunit() && !(u->getcarried())) move_land++;
             if (u->id() == TRANSPORTER) move_ferries++;
             if ((u->id() == UBOOT) ||
                 (u->id() == SCHNELLBOOT) ||
                 (u->id() == FREGATTE) ||
                 (u->id() == SCHLACHTSCHIFF)) move_ships++;
             if (u->id() == FLIEGER) move_air++;
             }
         u = (unit *) u->next();
         }

    land = new plan();

    if (move_air) move_planes(land);

    if (move_land) move_land_units(land);

    change_production(land);

    if (move_ferries) move_transporters(land);

    if (move_ships) move_navy(land);

    delete land;
    }

void game::move_planes(plan *land)
{   unit *u, *nextu;
    coordinate upos;
    int has_moved;

    land->settargets(AIRTARGETS);
    land->calcair();
  
    has_moved = 1;

    while(has_moved)
    {   has_moved = 0;

        u = (unit *) FIRSTMEMBER(units);
        while(u)
        {   nextu = (unit *) u->next();
            if (u->restmoves() && (u->id() == FLIEGER) && 
                (u->getowner(1) == activeplayer))
               if (!((flieger *) u)->should_return())
               {    upos = land->select_target_field(u->getpos());
                    if (upos != u->getpos())
                    {   u->movetowards(upos);
                        if (upos == u->getpos()) has_moved = 1;
                        nextu = (unit *) u->next();
                        u->alive();
                        }
                    }
            u = nextu;
            }
        }

    land->clearmap();
    land->settargets(HOMECITIES);
    land->calcair();

    has_moved = 1;

    while(has_moved)
    {   has_moved = 0;

        u = (unit *) FIRSTMEMBER(units);
        while(u)
        {   nextu = (unit *) u->next();
            if (u->restmoves() && (u->id() == FLIEGER) &&
                (u->getowner(1) == activeplayer))
               if (((flieger *) u)->should_return())
               {    upos = land->select_target_field(u->getpos());
                    if (upos != u->getpos())
                    {   u->movetowards(upos);
                        if (upos == u->getpos()) has_moved = 1;
                        nextu = (unit *) u->next();
                        u->alive();
                        }
                    }
            u = nextu;
            }
        }
    }

void game::move_land_units(plan *land)    
{   unit *u, *nextu;
    coordinate upos;
    int has_moved;

    land->clearmap();
    land->settargets(LANDTARGETS);
    land->calculate(2);
    land->addrandom();

    has_moved = 1;

    /* LOOP THROUGH ALL LAND UNITS. IF UNITS MOVE, LOOP AGAIN,
       UNTIL NO UNIT CAN MOVE ANY MORE */

    while(has_moved)
    {   has_moved = 0;

        u = (unit *) FIRSTMEMBER(units);
        while(u)
        {   nextu = (unit *) u->next();
            if (u->restmoves() && u->is_landunit() && (u->getowner(1) == activeplayer)
                  && !(u->getcarried()))
            {    upos = land->select_target_field(u->getpos());
                 if (upos != u->getpos())
                 {   u->movetowards(upos);
                     if (upos == u->getpos()) has_moved = 1;
                     nextu = (unit *) u->next();
                     u->alive();
                     }
                 }
            u = nextu;
            }
        }
    }

void game::change_production(plan *land)
{   city *c;

    /* SET PRODUCTION IN ALL CITIES (INCLUDING NEWLY CONQUERED ONES) */

    c = (city *) FIRSTMEMBER(g->cities);
    while(c)
    {   if ((c->getowner() == activeplayer) && 
            (!(c->produced) || (c->prodtime() == 2)))
        {    if (!(*land)[c->getpos()]) 
             {   if (coast(c->getpos())) c->setprod(TRANSPORTER);
                 else c->setprod(FLIEGER);
                 }
             else c->setprod(PANZER);
             }
        c = (city *) c->next();
        }
    }

void game::move_transporters(plan *land)
{   unit *u, *nextu;
    coordinate upos;
    int has_moved;

    land->clearmap();
    land->settargets(WATERTARGETS);
    land->calculate(0);
    land->blankblockedfields();
    land->calcwater();
    land->addrandom();

    has_moved = 1;

    // move full transporters:
    while(has_moved)
    {    has_moved = 0;

         u = (unit *) FIRSTMEMBER(units);
         while(u)
         {   nextu = (unit *) u->next();
             if ((u->id() == TRANSPORTER) && (u->getowner(1) == activeplayer)
                  && (u->restmoves()) && (((transporter *) u)->getnextunload()))
             {    upos = land->select_target_field(u->getpos());
                  if (upos != u->getpos())
                  {   u->movetowards(upos);
                      if (upos == u->getpos()) has_moved = 1;
                      nextu = (unit *) u->next();
                      u->alive();
                      }
                  }
             u = nextu;
             }
         }

    land->markownisles();
    land->clearmap();
    land->calculate(2);
    land->calcwater();
    land->addrandom();

    has_moved = 1;

    //move empty transporters:
    while(has_moved)
    {    has_moved = 0;

         u = (unit *) FIRSTMEMBER(units);
         while(u)
         {   nextu = (unit *) u->next();
             if ((u->id() == TRANSPORTER) && (u->getowner(1) == activeplayer)
                && !(((transporter *) u)->getnextunload()) && (u->restmoves()))
             {    upos = land->select_target_field(u->getpos(), 1);
                  if (upos != u->getpos())
                  {   u->movetowards(upos);
                      if (upos == u->getpos()) has_moved = 1;
                      nextu = (unit *) u->next();
                      u->alive();
                      }
                  }
             u = nextu;
             }
         }
    }

void game::move_navy(plan *land)
{   unit *u, *nextu;
    coordinate upos;
    int has_moved;

    land->clearmap();
    land->settargets(NAVALTARGETS);
    land->calcwater();
    land->addrandom();

    has_moved = 1;

    while(has_moved)
    {   has_moved = 0;

        u = (unit *) FIRSTMEMBER(units);
        while(u)
        {   nextu = (unit *) u->next();
            if ((u->getowner(1) == activeplayer) && (u->restmoves()) &&
                ((u->id() == SCHNELLBOOT) || (u->id() == UBOOT) || 
                (u->id() == FREGATTE) || (u->id() == SCHLACHTSCHIFF)))

            {    upos = land->select_target_field(u->getpos(), 1);
                 if (upos != u->getpos())
                 {   u->movetowards(upos);
                     if (upos == u->getpos()) has_moved = 1;
                     nextu = (unit *) u->next();
                     u->alive();
                     }
                 }
            u = nextu;
            }
        }
    }

//***************************************************** plan ************

plan::plan(void) : matrix()
{   coordinate i;

    linksoben.set(SX-1, SY-1);
    rechtsunten.set(0, 0);
    
    // find explored frame

    LOOP(i, 0, 0, SX-1, SY-1, 1)
       if (g->bekannt(i))
       {   if (i.x < linksoben.x) linksoben.x = i.x;
           if (i.x > rechtsunten.x) rechtsunten.x = i.x;
           if (i.y < linksoben.y) linksoben.y = i.y;
           if (i.y > rechtsunten.y) rechtsunten.y = i.y;
           }
   
    // add a five-square frame for planes flying blindly into the darkness

    linksoben.x -= 5;
    linksoben.y -= 5;
   
    rechtsunten.x += 5;
    rechtsunten.y += 5;

    // fit on the map:

    if (linksoben.x < 0) linksoben.x = 0;
    if (linksoben.y < 0) linksoben.y = 0;
    if (rechtsunten.x > SX-1) rechtsunten.x = SX -1;
    if (rechtsunten.y > SY-1) rechtsunten.y = SY -1;
    }

void plan::calculate(int add)
{   coordinate i, u;
    int found;

    do
    {   found = 0;
        LOOP(i, linksoben.x, linksoben.y, rechtsunten.x, rechtsunten.y, 1)
           if ((*this)[i])
               LOOP(u, i.x-1, i.y-1, i.x+1, i.y+1, 1)
                   if (ONMAP(u) && !(i == u) && !g->wasser(u) && (!((*this)[u]) || ((*this)[u] > (*this)[i]+add)))
                   {    (*this)[u] = (*this)[i]+add;
                        found = 1;
                        }
        }
    while (found);
    }

void plan::calcair(void)
{   coordinate i, u;
    int found;

    do
    {   found = 0;
        LOOP(i, linksoben.x, linksoben.y, rechtsunten.x, rechtsunten.y, 1)
           if ((*this)[i])
               LOOP(u, i.x-1, i.y-1, i.x+1, i.y+1, 1)
                   if (ONMAP(u) && !(i == u) && ((*this)[i] < 11) &&
                       (!((*this)[u]) || ((*this)[u] > (*this)[i]+1)))
                   {    (*this)[u] = (*this)[i]+1;
                        found = 1;
                        }
        }
    while (found);
    }

void plan::settargets(int mode)
{   coordinate i;

    LOOP(i, linksoben.x, linksoben.y, rechtsunten.x, rechtsunten.y, 1)
    {   switch(mode)
        {   case AIRTARGETS: // airplanes moving away

                 // primary goal: fight enemies
                 if (!(g->stadt(i)) && g->feindgesichtet(i) && (g->unitmap[i] != UBOOT))
                    (*this)[i] = 1;

                 // primary goal: explore
                 if (!(g->bekannt(i))) 
                    (*this)[i] = 1;

                 break;

            case HOMECITIES: // airplanes moving back
            
                 if (g->meins(i) && g->stadt(i))
                    (*this)[i] = 1;
      
                 break;

            case LANDTARGETS: // land units

                 // primary goal: conquer cities:
                 if (g->bekannt(i) && g->fremde_stadt(i))
                    (*this)[i] = 1;

                 // secondary goal: fight enemies:
                 if (!g->wasser(i) && !g->stadt(i) && g->feindgesichtet(i))
                    (*this)[i] = 5;

                 // tertiary goal: explore
                 if (!(g->bekannt(i))) 
                    (*this)[i] = 11;

                 // if nothing else is to do: board own transporters:
                 if (g->meins(i) && (g->unitmap[i] == TRANSPORTER))
                    (*this)[i] = 31;

                 break;

            case WATERTARGETS:  // full transporters
                 
                 //primary goal: islands with enemy units or cities
                 if (!g->wasser(i) && g->bekannt(i) &&
                     (g->feindgesichtet(i) || g->fremde_stadt(i)))
                    (*this)[i] = 1;

                 // secondary goal: explore
                 if (!(g->bekannt(i)))
                    (*this)[i] = 11;

                 break;

            case NAVALTARGETS: // battle ships and the like

                 if (g->wasser(i) && g->feindgesichtet(i) && (g->unitmap[i] != UBOOT))
                    (*this)[i] = 1;
                 
                 break;
            }
        }
    }
        
void plan::addrandom(void)
{   coordinate i;

    LOOP(i, linksoben.x, linksoben.y, rechtsunten.x, rechtsunten.y, 1)
        if ((*this)[i]) (*this)[i] += RANDOM(2);
    }

void plan::markownisles(void)
{   coordinate i;

    LOOP(i, linksoben.x, linksoben.y, rechtsunten.x, rechtsunten.y, 1)
       if (!((*this)[i]) && g->meins(i)) (*this)[i] = 220;
    }

void plan::blankblockedfields(void)
{    // blank out fields that cannot be used for transporter unload
     coordinate i;
     LOOP(i, linksoben.x, linksoben.y, rechtsunten.x, rechtsunten.y, 1)
        if (g->feindgesichtet(i)) (*this)[i] = 0;
     }

void plan::calcwater(void)
{   coordinate i, u;
    int found, add;

    do
    {   found = 0;
        LOOP(i, linksoben.x, linksoben.y, rechtsunten.x, rechtsunten.y, 1)
           if ((*this)[i] && ((*this)[i] < 220))
               LOOP(u, i.x-1, i.y-1, i.x+1, i.y+1, 1)
               if (ONMAP(u) && !(i == u) && g->wasser(u) && !g->minebekannt(u))
               {   add = 3;
                   if ((*(g->mp))[u] != 2) add = 4;
                   if (!((*this)[u]) || ((*this)[u] > (*this)[i]+add))
                   {    (*this)[u] = (*this)[i]+add;
                        found = 1;
                        }
                   }
        }
    while (found);
    }

void plan::clearmap(void)
// clears the map to 0, also prepares targets for empty transporters
// if own isles are set to 220 
{   coordinate i;

    LOOP(i, 0, 0, SX-1, SY-1, 1)
    {   if (((*this)[i] >= 220) && g->meins(i)) (*this)[i] = 1;
        else (*this)[i] = 0;
        }
    }

coordinate plan::select_target_field(coordinate upos, int water_only = 0)
{   coordinate m, i;
    int min, sign;

    m.set(-1,-1);
    min = 1000;
    sign = 1 - 2* RANDOM(2);

    SLOOP(i, upos.x - sign, upos.y - sign, upos.x + sign, upos.y + sign, sign)
        if (ONMAP(i) && !(i == upos) && (*this)[i] && ((*this)[i] < min))
        {    if (water_only && !(g->wasser(i))) continue;
             m = i;
             min = (*this)[i];
             }

    if (min < 1000)  
       return m;
    else 
       return upos;
    }
